package ru.dbotthepony.kstarbound.math

import com.google.gson.JsonArray
import ru.dbotthepony.kstarbound.api.IStruct2i
import ru.dbotthepony.kstarbound.api.IStruct3f
import ru.dbotthepony.kstarbound.api.IStruct4f
import kotlin.math.cos
import kotlin.math.sin

abstract class IVector2i<T : IVector2i<T>> : IMatrixLike, IMatrixLikeInt, IStruct2i {
	override val columns = 1
	override val rows = 2

	abstract val x: Int
	abstract val y: Int

	operator fun plus(other: IVector2i<*>) = make(x + other.x, y + other.y)
	operator fun plus(other: Int) = make(x + other, y + other)
	operator fun minus(other: IVector2i<*>) = make(x - other.x, y - other.y)
	operator fun minus(other: Int) = make(x - other, y - other)
	operator fun times(other: IVector2i<*>) = make(x * other.x, y * other.y)
	operator fun times(other: Int) = make(x * other, y * other)
	operator fun div(other: IVector2i<*>) = make(x / other.x, y / other.y)
	operator fun div(other: Int) = make(x / other, y / other)

	fun left() = make(x - 1, y)
	fun right() = make(x + 1, y)
	fun up() = make(x, y + 1)
	fun down() = make(x, y - 1)

	override fun get(row: Int, column: Int): Int {
		if (column != 0) {
			throw IndexOutOfBoundsException("Column must be 0 ($column given)")
		}

		return when (row) {
			0 -> x
			1 -> y
			else -> throw IndexOutOfBoundsException("Row out of bounds: $row")
		}
	}

	protected abstract fun make(x: Int, y: Int): T
}

data class Vector2i(override val x: Int = 0, override val y: Int = 0) : IVector2i<Vector2i>() {
	override fun make(x: Int, y: Int) = Vector2i(x, y)

	companion object {
		fun fromJson(input: JsonArray): Vector2i {
			return Vector2i(input[0].asInt, input[1].asInt)
		}

		val ZERO = Vector2i()
		val LEFT = Vector2i().left()
		val RIGHT = Vector2i().right()
		val UP = Vector2i().up()
		val DOWN = Vector2i().down()
	}
}

abstract class IVector3f<T : IVector3f<T>> : IMatrixLike, IMatrixLikeFloat, IStruct3f {
	override val columns = 1
	override val rows = 3

	abstract val x: Float
	abstract val y: Float
	abstract val z: Float

	operator fun plus(other: IVector3f<*>) = make(x + other.x, y + other.y, z + other.z)
	operator fun minus(other: IVector3f<*>) = make(x - other.x, y - other.y, z - other.z)
	operator fun times(other: IVector3f<*>) = make(x * other.x, y * other.y, z * other.z)
	operator fun div(other: IVector3f<*>) = make(x / other.x, y / other.y, z / other.z)

	operator fun plus(other: Float) = make(x + other, y + other, z + other)
	operator fun minus(other: Float) = make(x - other, y - other, z - other)
	operator fun times(other: Float) = make(x * other, y * other, z * other)
	operator fun div(other: Float) = make(x / other, y / other, z / other)

	override fun get(row: Int, column: Int): Float {
		if (column != 0) {
			throw IndexOutOfBoundsException("Column must be 0 ($column given)")
		}

		return when (row) {
			0 -> x
			1 -> y
			2 -> z
			else -> throw IndexOutOfBoundsException("Row out of bounds: $row")
		}
	}

	fun rotateAroundThis(rotation: Double): Matrix4f {
		val c = cos(rotation).toFloat()
		val s = sin(rotation).toFloat()
		val cInv = 1f - c

		return Matrix4f(
			m00 = c + x * x * cInv,			m01 = x * y * cInv - z * s,		m02 = x * z * cInv + y * s,
			m10 = y * x * cInv + z * s,		m11 = c + y * y * cInv,			m12 = y * z * cInv - x * s,
			m20 = z * x * cInv - y * s,		m21 = z * y * cInv + x * s,		m22 = c + z * z * cInv,
		)
	}

	operator fun times(other: IMatrixLikeFloat): T {
		if (other.rows >= 4 && other.columns >= 4) {
			val x = this.x * other[0, 0] +
					this.y * other[0, 1] +
					this.z * other[0, 2] +
					other[0, 3]

			val y = this.x * other[1, 0] +
					this.y * other[1, 1] +
					this.z * other[1, 2] +
					other[1, 3]

			val z = this.x * other[2, 0] +
					this.y * other[2, 1] +
					this.z * other[2, 2] +
					other[2, 3]

			return make(x, y, z)
		} else if (other.rows >= 3 && other.columns >= 3) {
			val x = this.x * other[0, 0] +
					this.y * other[0, 1] +
					this.z * other[0, 2]

			val y = this.x * other[1, 0] +
					this.y * other[1, 1] +
					this.z * other[1, 2]

			val z = this.x * other[2, 0] +
					this.y * other[2, 1] +
					this.z * other[2, 2]

			return make(x, y, z)
		}

		throw IllegalArgumentException("Incompatible matrix provided: ${other.rows} x ${other.columns}")
	}

	protected abstract fun make(x: Float, y: Float, z: Float): T
}

data class Vector3f(override val x: Float = 0f, override val y: Float = 0f, override val z: Float = 0f) : IVector3f<Vector3f>() {
	override fun make(x: Float, y: Float, z: Float): Vector3f {
		return Vector3f(x, y, z)
	}

	fun toMutableVector(): MutableVector3f {
		return MutableVector3f(x, y, z)
	}

	companion object {
		val UP = Vector3f(0f, 1f, 0f)
		val DOWN = Vector3f(0f, -1f, 0f)
		val LEFT = Vector3f(-1f, 0f, 0f)
		val RIGHT = Vector3f(1f, 0f, 0f)
		val FORWARD = Vector3f(0f, 0f, 1f)
		val BACKWARD = Vector3f(0f, 0f, -1f)
	}
}

data class MutableVector3f(override var x: Float = 0f, override var y: Float = 0f, override var z: Float = 0f) : IVector3f<MutableVector3f>() {
	fun toVector(): Vector3f {
		return Vector3f(x, y, z)
	}

	override fun make(x: Float, y: Float, z: Float): MutableVector3f {
		this.x = x
		this.y = y
		this.z = z
		return this
	}
}

abstract class IVector4f<T : IVector4f<T>> : IMatrixLike, IMatrixLikeFloat, IStruct4f {
	abstract val x: Float
	abstract val y: Float
	abstract val z: Float
	abstract val w: Float

	operator fun plus(other: IVector4f<*>) = make(x + other.x, y + other.y, z + other.z, w + other.w)
	operator fun minus(other: IVector4f<*>) = make(x - other.x, y - other.y, z - other.z, w + other.w)
	operator fun times(other: IVector4f<*>) = make(x * other.x, y * other.y, z * other.z, w + other.w)
	operator fun div(other: IVector4f<*>) = make(x / other.x, y / other.y, z / other.z, w + other.w)

	operator fun plus(other: Float) = make(x + other, y + other, z + other, w + other)
	operator fun minus(other: Float) = make(x - other, y - other, z - other, w - other)
	operator fun times(other: Float) = make(x * other, y * other, z * other, w * other)
	operator fun div(other: Float) = make(x / other, y / other, z / other, w / other)

	override val columns = 1
	override val rows = 4

	override fun get(row: Int, column: Int): Float {
		if (column != 0) {
			throw IndexOutOfBoundsException("Column must be 0 ($column given)")
		}

		return when (row) {
			0 -> x
			1 -> y
			2 -> z
			3 -> w
			else -> throw IndexOutOfBoundsException("Row out of bounds: $row")
		}
	}

	operator fun times(other: IMatrixLikeFloat): T {
		if (other.rows >= 4 && other.columns >= 4) {
			val x = this.x * other[0, 0] +
					this.y * other[0, 1] +
					this.z * other[0, 2] +
					this.w * other[0, 3]

			val y = this.x * other[1, 0] +
					this.y * other[1, 1] +
					this.z * other[1, 2] +
					this.w * other[1, 3]

			val z = this.x * other[2, 0] +
					this.y * other[2, 1] +
					this.z * other[2, 2] +
					this.w * other[2, 3]

			val w = this.x * other[3, 0] +
					this.y * other[3, 1] +
					this.z * other[3, 2] +
					this.w * other[3, 3]

			return make(x, y, z, w)
		}

		throw IllegalArgumentException("Incompatible matrix provided: ${other.rows} x ${other.columns}")
	}

	protected abstract fun make(x: Float, y: Float, z: Float, w: Float): T
}

data class Vector4f(override val x: Float = 0f, override val y: Float = 0f, override val z: Float = 0f, override val w: Float = 0f) : IVector4f<Vector4f>() {
	override fun make(x: Float, y: Float, z: Float, w: Float): Vector4f {
		return Vector4f(x, y, z, w)
	}
}

data class MutableVector4f(override var x: Float = 0f, override var y: Float = 0f, override var z: Float = 0f, override var w: Float = 0f) : IVector4f<MutableVector4f>() {
	override fun make(x: Float, y: Float, z: Float, w: Float): MutableVector4f {
		this.x = x
		this.y = y
		this.z = z
		this.w = w
		return this
	}
}
